#!/usr/bin/env python3

""" Based on https://github.com/DiamondLightSource/python-relion-yolo-it by
Sjors H.W. Scheres, Takanori Nakane, Colin M. Palmer, Donovan Webb"""

import argparse
import os
import shutil
import time
import subprocess
import math
from emtable import Table  # requires pip install emtable


RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"
DONE_MICS = "done_mics.txt"
CONDA_ENV = ". /home/gsharov/rc/conda.rc && conda activate topaz-0.2.4"
TOPAZ_PREPROCESS = "topaz preprocess"
TOPAZ_EXTRACT = "topaz extract"
TOPAZ_CONVERT = "topaz convert"
TOPAZ_SPLIT = "topaz split"
DEBUG = 1


def run_job(project_dir, args):
    start = time.time()
    in_mics = args.in_mics
    job_dir = args.out_dir
    thresh = args.threshold
    diam = args.diam
    model = args.model
    gpu = args.gpu
    threads = args.threads
    workers = args.workers

    getPath = lambda *arglist: os.path.join(project_dir, *arglist)

    if model != "None":
        model = getPath(model)

    # Reading the micrographs star file from relion
    optics = Table(fileName=getPath(in_mics), tableName='optics')
    angpix = float(optics[0].rlnMicrographPixelSize)
    mictable = Table(fileName=getPath(in_mics), tableName='micrographs')

    # calculate downscale factor to get to 4-8 A/px
    for scale in range(2, 20):
        angpix_bin = angpix * scale
        if diam < 300 and (4.0 <= angpix_bin <= 6.0 or angpix > 6.0):
            break
        elif diam >= 300 and (6.0 <= angpix_bin <= 8.0 or angpix > 8.0):
            break

    print("Using downscale factor %d to get %0.2f A/pix for %d A particle" % (
        scale, angpix_bin, diam))

    # Arranging files for topaz: making symlinks for mics
    done_mics = []
    mic_dirs = []
    if os.path.exists(DONE_MICS):
        with open(DONE_MICS, "r") as f:
            done_mics.extend(f.read().splitlines())
    if DEBUG:
        print("Current done_mics: ", done_mics)

    mic_fns = mictable.getColumnValues("rlnMicrographName")
    mic_ext = os.path.splitext(mic_fns[0])[1]
    input_job = "/".join(mic_fns[0].split("/")[:2])
    keys = ["/".join(i.split("/")[2:]) for i in mic_fns]  # remove JobType/jobXXX
    values = [os.path.splitext(i)[0] + "_topaz.star" for i in keys]  # _topaz.star
    mic_dict = {k: v for k, v in zip(keys, values)}

    for mic in mic_dict:
        if DEBUG:
            print("Processing mic: ", mic)
        mic_dir = os.path.dirname(mic)
        if not os.path.isdir(mic_dir):
            os.makedirs(mic_dir)
        if mic_dir not in mic_dirs:
            mic_dirs.append(mic_dir)
            print("Added folder %s to the mic_dirs" % mic_dir)
        if mic not in done_mics:
            inputfn = getPath(input_job, mic)
            outfn = getPath(job_dir, mic)

            # if mic had no coords picked, symlink still exists
            if os.path.exists(outfn):
                os.remove(outfn)
                continue
            else:
                os.symlink(inputfn, outfn)
                if DEBUG:
                    print("Link %s --> %s" % (inputfn, outfn))

    if len(mic_dict.keys()) == len(done_mics):
        print("All mics picked! Nothing to do.")
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()
        exit(0)

    # Launching topaz preprocess
    args_dict = {
        '--scale': scale,
        '--destdir': 'preprocessed',
        '--num-workers': workers,
        '--num-threads': threads
    }
    cmd = "%s && %s " % (CONDA_ENV, TOPAZ_PREPROCESS)
    cmd += " ".join(['%s %s' % (k, v) for k, v in args_dict.items()])
    for i in mic_dirs:
        cmd += " %s/*%s" % (i, mic_ext)

    print("Running command:\n{}".format(cmd))
    proc = subprocess.Popen(cmd, shell=True)
    proc.communicate()

    if proc.returncode:
        raise Exception("Command failed with return code %d" % proc.returncode)

    # Launching topaz extract
    args_dict = {
        '--radius': int(diam / angpix_bin // 2),
        '--up-scale': scale,
        '--threshold': thresh,
        '--output': 'output/coords.txt',
        '--num-workers': workers,
        '--num-threads': threads,
        '--device': gpu
    }

    if model != "None":
        args_dict.update({'--model': model})

    cmd = "%s && %s " % (CONDA_ENV, TOPAZ_EXTRACT)
    cmd += " ".join(['%s %s' % (k, v) for k, v in args_dict.items()])
    cmd += " preprocessed/*.mrc"
    os.makedirs("output", exist_ok=True)

    print("Running command:\n{}".format(cmd))
    proc = subprocess.Popen(cmd, shell=True)
    proc.communicate()

    if proc.returncode:
        raise Exception("Command failed with return code %d" % proc.returncode)

    # clean preprocessed dir
    shutil.rmtree("preprocessed")

    # Launching topaz convert
    cmd = "%s && %s " % (CONDA_ENV, TOPAZ_CONVERT)
    cmd += " -o output/coords.star output/coords.txt"
    print("Running command:\n{}".format(cmd))
    proc = subprocess.Popen(cmd, shell=True)
    proc.communicate()

    if proc.returncode:
        raise Exception("Command failed with return code %d" % proc.returncode)

    # Launching topaz split
    cmd = "%s && %s " % (CONDA_ENV, TOPAZ_SPLIT)
    cmd += " --output output output/coords.star"
    print("Running command:\n{}".format(cmd))
    proc = subprocess.Popen(cmd, shell=True)
    proc.communicate()

    if proc.returncode:
        raise Exception("Command failed with return code %d" % proc.returncode)

    # Move output star files for Relion to use
    with open(DONE_MICS, "a+") as f:
        for mic in mic_dict:
            mic_base = os.path.basename(mic)
            coord_topaz = "output/" + os.path.splitext(mic_base)[0] + ".star"
            coord_relion = mic_dict[mic]
            # To ensure that topaz doesn't pick from already done mics
            if os.path.exists(coord_topaz):
                f.write("%s\n" % mic)
                os.rename(coord_topaz, getPath(job_dir, coord_relion))
                os.remove(mic)  # clean up
                if DEBUG:
                    print("Moved %s to %s" % (coord_topaz, getPath(job_dir, coord_relion)))

    # clean output dir
    shutil.rmtree("output")

    # Required output mics star file
    with open("coords_suffix_topaz.star", "w") as mics_star:
        mics_star.write(in_mics)

    # Required output nodes star file
    nodes = Table(columns=['rlnPipeLineNodeName', 'rlnPipeLineNodeType'])
    nodes.addRow(os.path.join(job_dir, "coords_suffix_topaz.star"), "2")
    with open("RELION_OUTPUT_NODES.star", "w") as nodes_star:
        nodes.writeStar(nodes_star, tableName="output_nodes")

    outputFn = getPath(job_dir, "output_for_relion.star")
    if not os.path.exists(outputFn):
        # calculate diameter, original (boxSize) and downsampled (boxSizeSmall) box
        optics = Table(fileName=getPath(in_mics), tableName='optics')
        angpix = float(optics[0].rlnMicrographPixelSize)
        # use +10% for box size, make it even
        boxSize = math.ceil(diam * 1.1 / angpix / 2.) * 2

        # from relion_it.py script
        # Authors: Sjors H.W. Scheres, Takanori Nakane & Colin M. Palmer
        boxSizeSmall = None
        for box in (32, 48, 64, 96, 128, 160, 192, 256, 288, 300, 320,
                    360, 384, 400, 420, 450, 480, 512, 640, 768,
                    896, 1024):
            # Don't go larger than the original box
            if box > boxSize:
                boxSizeSmall = boxSize
                break
            # If Nyquist freq. is better than 8.5 A, use this
            # downscaled box, otherwise continue to next size up
            small_box_angpix = angpix * boxSize / box
            if small_box_angpix < 4.25:
                boxSizeSmall = box
                break

        print("\nEstimated parameters:\n\tDiameter (A): %d\n\tBox size (px): %d\n"
              "\tBox size binned (px): %d" % (diam, boxSize, boxSizeSmall))

        # output all params into a star file
        tableTopaz = Table(columns=['rlnParticleDiameter',
                                    'rlnOriginalImageSize',
                                    'rlnImageSize'])
        tableTopaz.addRow(diam, boxSize, boxSizeSmall)
        with open(outputFn, "w") as f:
            tableTopaz.writeStar(f, tableName='picker')

    end = time.time()
    diff = end - start
    print("Job duration = %dh %dmin %dsec \n" % (diff//3600, diff//60 % 60, diff % 60))


def main():
    """Change to the job working directory, then call run_job()"""
    help = """
External job for calling topaz within Relion 3.1.0. Run it in the main Relion project directory, e.g.:
    external_job_topaz.py --o External/topaz_picking --in_mics CtfFind/job004/micrographs_ctf.star --diam 120 --threshold 0 --gpu 0
"""
    parser = argparse.ArgumentParser(usage=help)
    parser.add_argument("--in_mics", help="Input micrographs STAR file")
    parser.add_argument("--o", dest="out_dir", help="Output directory name")
    parser.add_argument("--j", dest="threads", help="Number of CPU threads (default = 1)", type=int, default=1)
    parser.add_argument("--workers", dest="workers", help="Number of worker processes (default = 1)", type=int, default=1)
    parser.add_argument("--diam", help="Particle diameter in A (default = 120)", type=int, default=120)
    parser.add_argument("--threshold", help="Threshold for picking (default = -6)", type=float, default=-6)
    parser.add_argument("--model", help="Topaz training model (if not specified default is used)", default="None")
    parser.add_argument("--gpu", help='GPU to use (default = 0)', default="0")
    parser.add_argument("--pipeline_control", help="Not used here. Required by relion")

    args = parser.parse_args()

    if args.in_mics is None or args.out_dir is None:
        print("Error: --in_mics and --o are required params!")
        exit(1)

    if not args.in_mics.endswith(".star"):
        print("Error: --in_mics must point to a micrographs star file")
        exit(1)

    project_dir = os.getcwd()
    os.makedirs(args.out_dir, exist_ok=True)
    os.chdir(args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        run_job(project_dir, args)
    except:
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


if __name__ == "__main__":
    main()



# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

import os
from glob import glob
import unittest

from ..config import *
from ..utils.misc import getUsername, setParticleSizes
from ..parser import Parser


class TestParser(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.epuDir = os.path.join(os.path.dirname(__file__), "../Metadata-examples/EPU")
        cls.semDir = os.path.join(os.path.dirname(__file__), "../Metadata-examples/SerialEM/")

    def test_xml(self):
        print("=" * 80, "\nTesting xml parser...")
        xmlFns = glob(os.path.join(self.epuDir, "*/%s" % PATTERN_EPU), recursive=True)
        xmlFns = [os.path.abspath(x) for x in xmlFns]

        for mdFn in xmlFns:
            mdPath = "/".join(mdFn.split("/")[:-4])
            model = Parser()
            model.setSoftware("EPU")
            self._runParser(model, mdPath, mdFn)

    def test_mdoc(self):
        print("=" * 80, "\nTesting mdoc parser...")
        mdocFns = glob(os.path.join(self.semDir, PATTERN_MDOC))
        mdocFns = [os.path.abspath(x) for x in mdocFns]

        for mdFn in mdocFns:
            mdPath = "/".join(mdFn.split("/")[:-1])
            model = Parser()
            model.setSoftware("SerialEM")
            self._runParser(model, mdPath, mdFn)

    def _runParser(self, model, mdPath, mdFn):
        model.setPipeline(DEF_PIPELINE)
        model.setMdPath(mdPath)
        model.setFn(mdFn)

        username, uid = getUsername(mdPath)
        model.setUser(username, uid)
        model.acqDict['User'] = model.getUser()

        print("\n\nInput params: ",
              model.getSoftware(),
              model.getMdPath(),
              model.getUser(),
              model.getPipeline())
        print("\nFiles found: %s\n" % mdFn)

        if model.getSoftware() == 'EPU':
            model.parseImgEpu(mdFn)
        else:  # SerialEM
            model.parseImgMdoc(mdFn)

        model.calcDose()
        model.guessDataDir(testmode=True)
        model.acqDict['Picker'] = DEF_PICKER
        setParticleSizes(model)

        print("\nFinal parameters:\n")
        for k, v in sorted(model.acqDict.items()):
            print(k, v)
        print('\n')



# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

import re
import os
import math
import time
from glob import iglob

from .config import *
from .utils import parseXml, parseMrc


class Parser:
    """ Main parser class. """
    def __init__(self):
        # set default values
        self.mdPath = METADATA_PATH
        self.prjPath = PROJECT_PATH
        self.software = DEF_SOFTWARE
        self.user = DEF_USER
        self.fn = None
        self.pipeline = DEF_PIPELINE
        self.picker = DEF_PICKER
        self.size = (0, 0)

        self.acqDict = {
            'Mode': 'Linear',
            'NumSubFrames': '0',
            'Dose': '0',
            'OpticalGroup': 'opticsGroup1',
            'PhasePlateUsed': 'false',
            'GainReference': 'None',
            'DefectFile': 'None',
            'MTF': 'None'
        }

    def setMdPath(self, path):
        self.mdPath = path

    def getMdPath(self):
        return self.mdPath

    def setPipeline(self, choice):
        self.pipeline = choice

    def getPipeline(self):
        return self.pipeline

    def setPicker(self, choice):
        self.picker = choice

    def getPicker(self):
        return self.picker

    def getSize(self):
        return self.size

    def setSize(self, *args):
        self.size = args

    def getPrjPath(self):
        return self.prjPath

    def setPrjPath(self, path):
        self.prjPath = path

    def getUser(self):
        return self.user

    def setUser(self, login, uid):
        self.user = (login, uid)

    def getSoftware(self):
        return self.software

    def setSoftware(self, soft):
        self.software = soft

    def setFn(self, fn):
        self.fn = fn

    def getFn(self):
        return self.fn

    def guessFn(self, prog="EPU"):
        regex = PATTERN_EPU if prog == "EPU" else PATTERN_MDOC

        if DEBUG:
            print("\nUsing regex: ", regex)

        files = iglob(os.path.join(self.getMdPath(), regex))
        img = next(files, None)

        return img

    def parseImgEpu(self, fn):
        acqDict = parseXml(fn) if fn.endswith("xml") else parseMrc(fn)
        self.acqDict.update(acqDict)

    def parseImgMdoc(self, fn):
        """ SerialEM is used only for K2/K3 cameras. """
        self.acqDict['Detector'] = 'EF-CCD'

        with open(fn, 'r') as fname:
            regex = re.compile(REGEX_MDOC_VAR)

            for line in fname:
                match = regex.match(line)
                if match and match.groupdict()['var'] in SERIALEM_PARAMS:
                    key = match.groupdict()['var']
                    self.acqDict[key] = match.groupdict()['value'].strip()

        try:
            # rename a few keys to match EPU
            # T = SerialEM: Acquired on Titan Krios D3593
            match = re.search("D[0-9]{3,4}", self.acqDict['T'])
            if match:
                value = match.group().replace('D', '')
                self.acqDict['MicroscopeID'] = value
                self.acqDict.pop('T')
                if value in SCOPE_DICT:
                    self.acqDict['Cs'] = str(SCOPE_DICT[value][1])

            self.acqDict['Dose'] = self.acqDict.pop('ExposureDose')
            self.acqDict['AppliedDefocus'] = self.acqDict.pop('TargetDefocus')
            self.acqDict['Mode'] = 'Super-resolution' if self.acqDict['Binning'] == '0.5' else 'Counting'
            if 'PhasePlateInserted' in self.acqDict:
                self.acqDict['PhasePlateUsed'] = self.acqDict.pop('PhasePlateInserted')
            self.acqDict.pop('Binning')
        except KeyError:
            pass

        if DEBUG:
            for k, v in sorted(self.acqDict.items()):
                print("%s = %s" % (k, v))

    def calcDose(self):
        """ Calculate dose rate per unbinned px. """
        numFr = int(self.acqDict['NumSubFrames'])
        dose_total = float(self.acqDict['Dose'])  # e/A^2
        exp = float(self.acqDict['ExposureTime'])  # s

        if self.acqDict['Mode'] == 'Super-resolution':
            pix = 2 * float(self.acqDict['PixelSpacing'])  # A
        else:
            pix = float(self.acqDict['PixelSpacing'])  # A

        if numFr:  # not 0
            dose_per_frame = dose_total / numFr  # e/A^2/frame
        else:
            dose_per_frame = 0
        dose_on_camera = dose_total * math.pow(pix, 2) / exp  # e/ubpx/s

        self.acqDict['DosePerFrame'] = str(dose_per_frame)
        self.acqDict['DoseOnCamera'] = str(dose_on_camera)

    def calcBox(self, picker):
        """ Calculate box, mask, downsample. """
        minSize, maxSize = self.acqDict['PtclSizes']
        ptclSizeAng = max(minSize, maxSize) if picker == 'LogPicker' else minSize
        angpix = float(self.acqDict['PixelSpacing'])

        if self.acqDict['Mode'] == 'Super-resolution':
            # since we always bin by 2 in mc if using super-res
            angpix *= 2

        ptclSizePx = float(ptclSizeAng) / angpix
        # use +20% for mask size
        self.acqDict['MaskSize'] = str(math.ceil(1.2 * ptclSizePx))
        # use +30% for box size, make it even
        boxSize = 1.3 * ptclSizePx
        self.acqDict['BoxSize'] = str(math.ceil(boxSize / 2.) * 2)

        # from relion_it.py script
        # Authors: Sjors H.W. Scheres, Takanori Nakane & Colin M. Palmer
        for box in (32, 48, 64, 96, 128, 160, 192, 256, 288, 300, 320,
                    360, 384, 400, 420, 450, 480, 512, 640, 768,
                    896, 1024):
            # Don't go larger than the original box
            if box > boxSize:
                self.acqDict['BoxSizeSmall'] = str(boxSize)
                break
            # If Nyquist freq. is better than 8.5 A, use this
            # downscaled box, otherwise continue to next size up
            small_box_angpix = angpix * boxSize / box
            if small_box_angpix < 4.25:
                self.acqDict['BoxSizeSmall'] = str(box)
                break

    def guessDataDir(self, wait=False, testmode=False):
        """ Guess folder name with movies, gain and defects files. """
        movieDir, gainFn, defFn = 'None', 'None', 'None'
        camera = self.acqDict['Detector']
        scopeID = self.acqDict['MicroscopeID']

        # get MTF file
        if camera == 'EF-CCD':
            model = SCOPE_DICT[scopeID][3]
            if model is not None:
                self.acqDict['MTF'] = MTF_DICT[model]
        else:
            model = SCOPE_DICT[scopeID][2]
            if self.acqDict['Mode'] == 'Linear':
                self.acqDict['MTF'] = MTF_DICT['%s-linear' % model]
            else:
                self.acqDict['MTF'] = MTF_DICT['%s-count' % model]

        # update with real camera name
        self.acqDict['Detector'] = model

        if self.getSoftware() == 'EPU':
            p1 = MOVIE_PATH_DICT[camera] % (SCOPE_DICT[scopeID][0], model)
            session = os.path.basename(self.getMdPath())

            if camera == 'EF-CCD':
                movieDir = os.path.join(p1, "DoseFractions", session, EPU_MOVIES_DICT[model])
                movieBaseDir = os.path.join(p1, "DoseFractions", session)
                gainFiles = iglob(os.path.join(os.path.dirname(movieDir), GAIN_DICT[model]))
                gainFn = next(gainFiles, 'None')
            else:
                movieDir = os.path.join(p1, session, EPU_MOVIES_DICT[model])
                movieBaseDir = os.path.join(p1, session)

            if wait:  # in daemon mode wait for movie folder to appear
                while True:
                    if not os.path.exists(movieBaseDir):
                        print("Movie folder %s not found, waiting for 1 min.." % movieBaseDir)
                        time.sleep(60)
                    else:
                        print("Movie folder found! Continuing..")
                        break
            else:  # GUI mode
                if not os.path.exists(movieBaseDir) and not testmode:
                    raise FileNotFoundError("Movie folder %s does not exist!" % movieBaseDir)

        else:  # SerialEM
            movieDir = os.path.join(self.getMdPath(), PATTERN_SEM_MOVIES)
            gainFn = os.path.join(self.getMdPath(), self.acqDict['GainReference'])
            defFn = os.path.join(self.getMdPath(), self.acqDict['DefectFile'])

        # populate dict
        self.acqDict['Software'] = self.getSoftware()
        self.acqDict['PrjPath'] = self.getPrjPath()
        self.acqDict['MoviePath'] = movieDir
        if os.path.exists(gainFn):
            self.acqDict['GainReference'] = gainFn
        if os.path.exists(defFn):
            self.acqDict['DefectFile'] = defFn



# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

import os
import shutil
import subprocess
import json

from .utils.misc import precalculateVars, getPrjName
from .config import DEBUG, JSON_TEMPLATE, SCHEDULE_PATH


def setupRelion(paramDict):
    """ Prepare and launch Relion 3.1 schedules. """
    bin, gain, defect, group_frames = precalculateVars(paramDict)
    mapDict = {'Cs': paramDict['Cs'],
               'dose_rate': paramDict['DosePerFrame'],
               'angpix': paramDict['PixelSpacing'],
               'voltage': paramDict['Voltage'],
               'motioncorr_bin': bin,
               'group_frames': group_frames,
               'is_VPP': paramDict['PhasePlateUsed'],
               'optics_group': paramDict['OpticalGroup'],
               'size_min': paramDict['PtclSizes'][0],
               'size_max': paramDict['PtclSizes'][1],
               }

    pickerSchedule = {'crYOLO': 'preprocess-cryolo',
                      'Topaz': 'preprocess-topaz',
                      'LogPicker': 'preprocess-logpicker'}
    preprocess_schd = pickerSchedule[paramDict['Picker']]

    if paramDict['PtclSizes'][0] != 0:
        mapDict.update({
            'box_size': paramDict['BoxSize'],
            'mask_diam_px': paramDict['MaskSize'],
            'mask_diam': int(paramDict['MaskSize']) * bin * float(paramDict['PixelSpacing']),
            'box_size_bin': paramDict['BoxSizeSmall'],
        })

    prjName = getPrjName(paramDict)
    prjPath = os.path.join(paramDict['PrjPath'], prjName)
    os.mkdir(prjPath)

    # Create links
    movieDir = os.path.join(prjPath, "Movies")
    if os.path.islink(movieDir):
        os.remove(movieDir)
    if os.path.exists(movieDir):
        raise FileExistsError('Destination %s already exists and is not a link' %
                              movieDir)

    if paramDict['Software'] == 'EPU':
        # EPU: Movies -> EPU session folder
        origPath1, origPath2 = paramDict['MoviePath'].split('/Images-Disc')
        mapDict['movies_wildcard'] = 'Movies/Images-Disc%s' % origPath2
    else:
        # SerialEM: Movies -> Raw path folder
        origPath1 = paramDict['MoviePath'].split('*.tif')[0]
        mapDict['movies_wildcard'] = 'Movies/*.tif'

    if DEBUG:
        print("Params passed to Relion: ", mapDict)

    os.symlink(origPath1, movieDir)
    os.chdir(prjPath)
    for i in [gain, defect, paramDict['MTF']]:
        if os.path.exists(i):
            shutil.copyfile(i, os.path.basename(i))

    mapDict['gainref'] = os.path.basename(gain) or '""'
    mapDict['mtf_file'] = os.path.basename(paramDict['MTF'])
    mapDict['defect_file'] = os.path.basename(defect) or '""'

    try:
        subprocess.check_output(["which", "relion_scheduler"],
                                stderr=subprocess.DEVNULL)
        if not os.path.exists('Schedules'):
            shutil.copytree(SCHEDULE_PATH, os.getcwd() + '/Schedules')
    except subprocess.CalledProcessError:
        print("ERROR: relion command not found in PATH or Schedules not found!")
        exit(1)

    # setup ACL for uid
    uid = paramDict['User'][1]
    if uid:  # not zero
        cmdList = ["setfacl -R -m u:%s:rwX %s" % (uid, prjPath)]
        cmdList.append("setfacl -R -d -m u:%s:rwX %s" % (uid, prjPath))
        cmdList.append("setfacl -R -m m::rwx %s" % prjPath)
        cmdList.append("setfacl -R -d -m m::rwx %s" % prjPath)
        try:
            for cmd in cmdList:
                subprocess.check_output(cmd.split())
        except subprocess.CalledProcessError:
            print("Warning: setfacl command failed, ignoring..")

    # Set up scheduler vars
    cmdList = list()
    for key in mapDict:
        cmd = 'relion_scheduler --schedule %s --set_var %s --value %s' % (
            preprocess_schd, key, str(mapDict[key]))
        cmdList.append(cmd)

    cmdList.append('relion_scheduler --schedule class2d --set_var mask_diam --value %s' % mapDict.get('mask_diam', 0))

    for cmd in cmdList:
        print(cmd)
        proc = subprocess.check_output(cmd.split())

    # Run scheduler with Popen - without waiting for return
    cmd1 = 'relion_scheduler --schedule %s --run --pipeline_control Schedules/%s/ &' % (
        preprocess_schd, preprocess_schd)
    cmd2 = 'relion_scheduler --schedule class2d --run --pipeline_control Schedules/class2d/ &'

    with open("schedules_preprocess.log", "w") as fnLog1:
        print(cmd1)
        proc = subprocess.Popen(cmd1.split(), universal_newlines=True,
                                stdout=fnLog1, stderr=subprocess.STDOUT)
    with open("schedules_class2d.log", "w") as fnLog2:
        print(cmd2)
        proc = subprocess.Popen(cmd2.split(), universal_newlines=True,
                                stdout=fnLog2, stderr=subprocess.STDOUT)


def setupScipion(paramDict):
    """ Prepare and schedule Scipion 3 workflow.
    The default template will run import movies, motioncor2,
    gctf, cryolo picking, extraction and summary monitor protocols.
    """
    prjName = getPrjName(paramDict)
    prjPath = os.path.join(paramDict['PrjPath'], prjName)
    os.mkdir(prjPath)
    os.chdir(prjPath)
    with open(JSON_TEMPLATE, 'r') as f:
        protocolsList = json.load(f)
    protNames = dict()

    for i, protDict in enumerate(protocolsList):
        protClassName = protDict['object.className']
        protNames[protClassName] = i

    bin, gain, defect, group_frames = precalculateVars(paramDict)
    for i in [gain, defect, paramDict['MTF']]:
        if os.path.exists(i):
            shutil.copyfile(i, os.path.basename(i))

    # import movies
    importProt = protocolsList[protNames["ProtImportMovies"]]

    if paramDict['Software'] == 'EPU':
        # get EPU session folder
        origPath1, origPath2 = paramDict['MoviePath'].split('Images-Disc')
        importProt["filesPath"] = origPath1
        importProt["filesPattern"] = "Images-Disc%s" % origPath2
    else:
        # SerialEM
        importProt["filesPath"] = "%s/" % "/".join(paramDict['MoviePath'].split('/')[:-1])
        importProt["filesPattern"] = paramDict['MoviePath'].split('/')[-1]

    importProt["voltage"] = float(paramDict['Voltage'])
    importProt["sphericalAberration"] = float(paramDict['Cs'])
    importProt["samplingRate"] = float(paramDict['PixelSpacing'])
    importProt["dosePerFrame"] = float(paramDict['DosePerFrame'])
    importProt["gainFile"] = os.path.join(os.getcwd(), os.path.basename(gain)) if gain else ""

    # motioncorr
    movieProt = protocolsList[protNames["ProtMotionCorr"]]
    movieProt["binFactor"] = bin
    movieProt["group"] = group_frames
    movieProt["defectFile"] = os.path.join(os.getcwd(), os.path.basename(defect)) if defect else ""

    # gctf
    ctfProt = protocolsList[protNames["ProtGctf"]]
    ctfProt["doPhShEst"] = paramDict['PhasePlateUsed']

    jsonFn = "%s.json" % prjName
    with open(jsonFn, "w") as f:
        jsonStr = json.dumps(protocolsList, indent=4, separators=(',', ': '))
        f.write(jsonStr)

    try:
        subprocess.check_output(["which", "scipion3"], stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        print("ERROR: scipion3 command not found in PATH!")
        exit(1)

    cmd = 'scipion3 python -m pyworkflow.project.scripts.create %s %s' % (
        prjName, os.path.abspath(jsonFn))
    proc = subprocess.run(cmd.split(), check=True)

    cmd2 = 'scipion3 python -m pyworkflow.project.scripts.schedule %s' % prjName
    proc2 = subprocess.Popen(cmd2.split(), universal_newlines=True)



# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

# Config variables
VERSION = "0.9.10"
DEBUG = 1

# daemon mode vars
DEF_USER = ("emuser", 11048)  # default name: (username, uid)
DEF_PIPELINE = "Relion"  # default pipeline: Relion or Scipion
DEF_PICKER = "crYOLO"  # default particle picker: crYOLO or Topaz or LogPicker
DEF_SOFTWARE = "EPU"  # default software: EPU or SerialEM
DEF_PREFIX = "lmb_"  # found metadata folder name should start with this prefix

LOGPICKER_SIZES = (150, 180)  # min/max size in A for Relion LogPicker
TOPAZ_SIZE = 150  # diameter in A for Topaz

# path to EPU session or folder with SerialEM mdoc files
METADATA_PATH = "/mnt/MetaData/Krios1/EPU/OTFP"
#METADATA_PATH = "/home/gsharov/soft/MDCatch/mdcatch/Metadata-examples/EPU"

# path where Relion projects are created
# for Scipion mtf, defects, gain and template files are copied here
PROJECT_PATH = "/cephfs"

# Folder with Relion 3.1 schedules
SCHEDULE_PATH = "/home/gsharov/soft/MDCatch/mdcatch/Schedules"

# Scipion pre-processing template
JSON_TEMPLATE = "/home/gsharov/soft/MDCatch/mdcatch/template.json"

# main dictionary
# instrumentID: [name, Cs, TFS camera, Gatan camera]
SCOPE_DICT = {'3299': ['Krios1', 2.7, 'Falcon', 'K2'],
              '3413': ['Krios2', 2.7, 'Falcon', 'K2'],
              '3593': ['Krios3', 2.7, 'Falcon', 'K3'],
              '9952833': ['Glacios', 2.7, 'Falcon', None]
              }

###############################################################################
# EPU params
EPU_MOVIES_DICT = {'Falcon': "Images-Disc*/GridSquare_*/Data/FoilHole_*_Data_*_Fractions.mrc",
                   'K2': "Images-Disc*/GridSquare_*/Data/FoilHole_*_Data_*.mrc",
                   'K3': "Images-Disc*/GridSquare_*/Data/FoilHole_*_Data_*_fractions.tiff"
                   }
GAIN_DICT = {'K2': "FoilHole_*_Data_*-gain-ref.MRC",
             'K3': "FoilHole_*_Data_*_gain.tiff"
             }

# change the pattern below if you want to parse movie sums mrc instead
PATTERN_EPU = "Images-Disc*/GridSquare_*/Data/FoilHole_*_Data_*.xml"

# path to MTF files for Relion (300 kV only)
# examples: Name-count, Name-linear or Name,
# camera names should match SCOPE_DICT
MTF_DICT = {
    'Falcon-count': '/home/gsharov/soft/MTFs/mtf_falcon3EC_300kV.star',
    'Falcon-linear': '/home/gsharov/soft/MTFs/mtf_falcon2_300kV.star',
    'K2': '/home/gsharov/soft/MTFs/mtf_k2_300kV.star',
    'K3': '/home/gsharov/soft/MTFs/mtf_K3_300kv_nocds.star'
}

# path to raw movies folder depending on camera name
# example: /mnt/Data/Krios3/K3/
# scope and camera names should match SCOPE_DICT
MOVIE_PATH_DICT = {
    'EF-CCD': '/mnt/Data/%s/%s/',
    'BM-Falcon': '/mnt/Data/%s/%s/'
}

###############################################################################
# SerialEM params
PATTERN_SEM_MOVIES = "*.tif"  # movie pattern for Gatan K2/K3 cameras
PATTERN_MDOC = "*.tif.mdoc"
REGEX_MDOC_VAR = "(?P<var>[a-zA-Z0-9]+?) = (?P<value>(.*))"

# SerialEM mdoc vars to parse
SERIALEM_PARAMS = [
    'T',  # Microscope ID
    'Voltage',
    'Magnification',
    'ExposureDose',
    'PixelSpacing',
    'ExposureTime',
    'Binning',
    'TargetDefocus',
    'NumSubFrames',
    # 'DateTime',
    'DefectFile',
    'GainReference',
    # optional vars below can be added to mdoc using "AddToNextFrameStackMdoc key value"
    'OpticalGroup',
    'PhasePlateInserted',
    'BeamTiltCompensation',
    'Beamtilt'
]

# help message for path selection
help_message = """Select the following folder:\n\n
   For EPU: the EPU session folder on /mnt/MetaData/
   with Images-DiscX folder inside.\n
   OR\n
   For SerialEM: the folder on /mnt/Data/ that
   contains tif movies and mdoc files.\n"""

# help message for particle size
help_picker = """Provide particle size in Angstroms.\n\n
   - crYOLO can estimate it automatically (set this to 0)\n
   - Topaz requires this parameter to remove neighboring picks\n
   - LogPicker requires Min and Max diameter"""



#!/usr/bin/env python3
# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

""" This script parses BigTIFF file header from SerialEM movies. """

from struct import unpack

from .dtypes import bigtiff_tags

file = '/home/azazello/test/parser/K3-serialem.tif'
header_dict = dict()

with open(file, 'rb') as fin:
    fin.seek(0)
    header = fin.read(16)
    try:
        byteorder = {b'II': '<', b'MM': '>'}[header[:2]]
    except KeyError:
        print('Not a TIFF file!')

    version = unpack(byteorder + 'H', header[2:4])[0]  # 43 = BigTIFF
    if version == 42:
        raise NotImplemented('Only BigTIFF files are supported!')

    ifd_offset = unpack(byteorder + 'Q', header[8:16])[0]  # offset to the first IFD
    # find number of tags in the first IFD
    fin.seek(ifd_offset)
    num_tags = unpack(byteorder + 'Q', fin.read(8))[0]
    start = int(ifd_offset) + 8

    for _ in range(1, num_tags+1):
        fin.seek(start)
        tag = fin.read(20)  # every tag takes 20 bytes
        tagnum = unpack(byteorder + 'H', tag[:2])[0]  # IFD tag
        tagname = bigtiff_tags[tagnum][0]
        dtype = unpack(byteorder + 'H', tag[2:4])[0]  # IFD type

        if dtype in [3, 16]:  # SHORT 2-byte uint / LONG8 8-byte uint
            header_dict[tagname] = unpack(byteorder + 'Q', tag[12:20])[0]  # IFD value
        elif dtype == 5:  # RATIONAL 2x 4-byte uint
            header_dict[tagname] = unpack(byteorder + 'L', tag[12:16])[0]
        elif dtype == 2:  # ASCII 8-byte
            count = unpack(byteorder + 'Q', tag[4:12])[0]
            offset = unpack(byteorder + 'Q', tag[12:20])[0]
            header_dict[tagname] = (count, offset)
        elif dtype == 1:  # BYTE 1-byte uint
            header_dict[tagname] = unpack(byteorder + 'B', tag[12:13])[0]
        start += 20

    # parse tags 270=ImageDescription, 306=DateTime && update dict
    for i in ["ImageDescription", "DateTime"]:
        count_desc, offset_desc = header_dict[i]
        fin.seek(offset_desc)
        part = fin.read(count_desc)
        res = unpack(byteorder + '%ds' % count_desc,
                     part[:count_desc+1])[0]
        header_dict[i] = res.decode('utf-8').strip('\x00')

    # count number of IFDs
    c = 0
    next_offset = ifd_offset
    while next_offset != 0:
        fin.seek(next_offset)
        num_tags = int(unpack(byteorder + 'Q', fin.read(8))[0])
        fin.seek(next_offset + num_tags * 20 + 8)
        next_offset = int(unpack(byteorder + 'Q', fin.read(8))[0])
        c += 1
    header_dict["nimg"] = c

for k, v in sorted(header_dict.items()):
    print("%s = %s" % (k, v))



# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

import os
import subprocess
from datetime import datetime

from ..config import (DEF_USER, SCOPE_DICT, DEF_PICKER,
                      TOPAZ_SIZE, LOGPICKER_SIZES)


def getUsername(mdPath):
    """ Return username and uid using NIS database. """
    mdFolder = os.path.basename(mdPath)
    try:
        username = mdFolder.split("_")[1]
    except IndexError:
        return DEF_USER

    cmd = "/usr/bin/ypmatch %s passwd" % username
    try:
        res = subprocess.check_output(cmd.split())
    except subprocess.CalledProcessError:
        print("ERROR: username %s not found! Using defaults: " % username, DEF_USER)
        return DEF_USER
    except FileNotFoundError:
        print("ERROR: command %s failed! Using defaults: " % cmd, DEF_USER)
        return DEF_USER
    else:
        uid = str(res).split(":")[2]
        return username, uid


def getPrjName(paramDict):
    """ Util func to get project name.
    Returns project name = username_scope_date_time
    """
    username, _ = paramDict['User']
    scope = SCOPE_DICT[paramDict['MicroscopeID']][0]
    dt = datetime.now()
    dt = dt.strftime('%d-%m-%Y_%H%M')
    prjName = username + '_' + scope + '_' + dt

    return prjName


def precalculateVars(paramDict):
    """ Returns binning, gain, defect files and frame grouping. """
    # set motioncor bin to 2 if using super-res data
    bin = 2.0 if paramDict['Mode'] == 'Super-resolution' else 1.0
    gain = '' if paramDict['GainReference'] == 'None' else paramDict['GainReference']
    defect = '' if paramDict['DefectFile'] == 'None' else paramDict['DefectFile']
    # group frames if dose rate < 0.8 e/A^2/frame
    group_frames = 1
    dpf = float(paramDict['DosePerFrame'])
    if dpf < 0.8:
        group_frames += 1
        if group_frames * dpf < 0.8:
            group_frames += 1

    return bin, gain, defect, group_frames


def setParticleSizes(model):
    """ Setup particle sizes when running in non-GUI mode. """
    if DEF_PICKER == 'Topaz':
        model.acqDict['PtclSizes'] = TOPAZ_SIZE, LOGPICKER_SIZES[1]
        model.calcBox(DEF_PICKER)
    elif DEF_PICKER == 'LogPicker':
        model.acqDict['PtclSizes'] = LOGPICKER_SIZES
        model.calcBox(DEF_PICKER)
    else:
        model.acqDict['PtclSizes'] = 0, LOGPICKER_SIZES[1]



# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

""" This script parses MRC 2014 FEI1 file header. """

import numpy as np
import math
from struct import unpack

from .dtypes import mrc_tags, fei_tags
from ..config import DEBUG, SCOPE_DICT


def parseMrc(fn):
    with open(fn, 'rb') as fin:
        fin.seek(0)
        header = fin.read(1024)
        map = header[208:212]
        machst = header[212:214]
        exttyp = header[104:108]

        if map != b'MAP ' or machst != b'DD':  # 0x44 0x44 is little endian
            print("Not a MRC file or not little endian byte order!")
            return None
        if exttyp != b'FEI1':
            print("No FEI1 extended header found!")
            return None

        fin.seek(1024)
        md_size = unpack('<L', fin.read(4))[0]
        fin.seek(1024)
        ext_header = fin.read(md_size)

    # parse main header
    header_arr = np.frombuffer(header[:52], dtype=mrc_tags)
    keys = header_arr.dtype.names
    header_dict = [dict(zip(keys, value)) for value in header_arr][0]

    header_dict["apix_x"] = header_dict["CELLA"]["X"] / header_dict["NX"]
    header_dict["apix_y"] = header_dict["CELLA"]["Y"] / header_dict["NY"]

    # parse extended header
    ext_header_arr = np.frombuffer(ext_header, dtype=fei_tags)
    keys = ext_header_arr.dtype.names
    ext_header_dict = [dict(zip(keys, value)) for value in ext_header_arr][0]

    acqDict = {**header_dict, **ext_header_dict}
    newDict = _standardizeDict(acqDict)

    if DEBUG:
        for k, v in sorted(newDict.items()):
            print("%s = %s" % (k, v))

    return newDict


def _standardizeDict(acqDict):
    """Convert values to expected format. """
    stdDict = {
        'AppliedDefocus': float(acqDict['Applied defocus']) * math.pow(10, 6),
        'BeamShiftX': acqDict['Shift X'],  # TEM beam-image shift
        'BeamShiftY': acqDict['Shift Y'],
        'ImageShiftX': acqDict['Shift offset X'],  # TEM pure image shift
        'ImageShiftY': acqDict['Shift offset Y'],
        'BeamSize': float(acqDict['Illuminated area']) * math.pow(10, 6),
        'Detector': acqDict['Camera name'].decode("utf-8"),
        'Dose': float(acqDict['Dose']) * math.pow(10, -20),
        'EPUversion': acqDict['Application version'].decode("utf-8"),
        'ExposureTime': acqDict['Integration time'],
        'Magnification': acqDict['Magnification'],
        'MicroscopeID': acqDict['D-Number'].decode("utf-8"),
        # no way to find super-res?
        'Mode': 'Counting' if bool(acqDict['Direct detector electron counting']) else 'Linear',
        # no way to find frames number?
        'NumSubFrames': 0,
        'SpotSize': acqDict['Spot index'],
        'Voltage': float(acqDict['HT']) // 1000,
        'PhasePlateUsed': bool(acqDict['Phase Plate']),
    }

    sr = 1
    stdDict['PixelSpacing'] = float(acqDict['Pixel size X']) * math.pow(10, 10) / sr

    if stdDict['MicroscopeID'] in SCOPE_DICT:
        stdDict['Cs'] = SCOPE_DICT[stdDict['MicroscopeID']][1]

    # convert all to str
    for key in stdDict:
        stdDict[key] = str(stdDict[key])

    return stdDict



#!/usr/bin/env python3

""" Based on https://github.com/DiamondLightSource/python-relion-yolo-it by
Sjors H.W. Scheres, Takanori Nakane, Colin M. Palmer, Donovan Webb"""

import argparse
import json
import os
import time
import subprocess
from emtable import Table  # requires pip install emtable


RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"
CONDA_ENV = ". /home/gsharov/rc/conda.rc && conda activate cryolo-1.7.5"
CRYOLO_TRAIN = "cryolo_train.py"
CRYOLO_GEN_MODEL = "/home/gsharov/soft/cryolo/gmodel_phosnet_202005_N63_c17.h5"
CRYOLO_JANNI_MODEL = "/home/gsharov/soft/cryolo/gmodel_janni_20190703.h5"
TUNE_MODEL = "fine_tuned_model.h5"
IMG_FOLDER = "train_image"
ANNOT_FOLDER = "train_annot"
DEBUG = 1


def run_job(project_dir, args):
    start = time.time()
    in_parts = args.in_parts
    job_dir = args.out_dir
    box_size = args.box_size
    model = args.model or CRYOLO_GEN_MODEL
    gpus = args.gpu

    getPath = lambda *arglist: os.path.join(project_dir, *arglist)

    # Making a cryolo config file
    json_dict = {
        "model": {
            "architecture": "PhosaurusNet",
            "input_size": 1024,
            "max_box_per_image": 600,
            "anchors": [box_size, box_size],
            "norm": "STANDARD",
            "filter": [
                0.1,
                "filtered_tmp/"
            ]
        },
        "train": {
            "train_image_folder": IMG_FOLDER,
            "train_annot_folder": ANNOT_FOLDER,
            "train_times": 10,
            "batch_size": 6,
            "learning_rate": 0.0001,
            "nb_epoch": 200,
            "object_scale": 5.0,
            "no_object_scale": 1.0,
            "coord_scale": 1.0,
            "class_scale": 1.0,
            "pretrained_weights": "%s" % model,
            "saved_weights_name": getPath(job_dir, TUNE_MODEL),
            "debug": True
        },
        "valid": {
            "valid_image_folder": "",
            "valid_annot_folder": "",
            "valid_times": 1
        }
    }

    if DEBUG:
        print("Using following config.json: ", json_dict)

    with open("config.json", "w") as json_file:
        json.dump(json_dict, json_file, indent=4)

    # Create folder structure for cryolo
    os.mkdir(IMG_FOLDER)
    os.mkdir(ANNOT_FOLDER)

    # Reading the particles star file from relion
    parttable = Table(fileName=getPath(in_parts), tableName='particles')
    mics_dict = {}

    # Arranging files for cryolo: making symlinks for mics and creating box files
    for row in parttable:
        mic = row.rlnMicrographName
        xCoord = int(int(row.rlnCoordinateX) - box_size / 2)
        yCoord = int(int(row.rlnCoordinateY) - box_size / 2)
        if mic in mics_dict:
            mics_dict[mic].append((xCoord, yCoord))
        else:
            mics_dict[mic] = [(xCoord, yCoord)]

    for mic in mics_dict:
        micSrc = getPath(mic)
        micDst = getPath(job_dir, IMG_FOLDER, os.path.basename(mic))
        if not os.path.exists(micDst):
            os.symlink(micSrc, micDst)
        if DEBUG:
            print("Link %s --> %s" % (micSrc, micDst))

        box = os.path.splitext(micDst)[0] + ".box"
        box = box.replace(IMG_FOLDER, ANNOT_FOLDER)
        with open(box, "w") as f:
            for coords in mics_dict[mic]:
                f.write("%s\t%s\t%s\t%s\n" %
                        (coords[0], coords[1], box_size, box_size))
        if DEBUG:
            print("Created box file: %s" % box)
    
    # Launching cryolo
    args_dict = {
        '--conf': "config.json",
        '--gpu': gpus,
        '--warmup': 0,
        '--fine_tune': "",
        '--cleanup': ""
    }
    cmd = "%s && %s " % (CONDA_ENV, CRYOLO_TRAIN)
    cmd += " ".join(['%s %s' % (k, v) for k, v in args_dict.items()])

    print("Running command:\n{}".format(cmd))
    proc = subprocess.Popen(cmd, shell=True)
    proc.communicate()

    if proc.returncode:
        raise Exception("Command failed with return code %d" % proc.returncode)

    # Required output star file
    with open("_manualpick.star", "w") as f:
        f.write(in_parts)

    # Required output nodes star file
    nodes = Table(columns=['rlnPipeLineNodeName', 'rlnPipeLineNodeType'])
    nodes.addRow(os.path.join(job_dir, "_manualpick.star"), "2")
    with open("RELION_OUTPUT_NODES.star", "w") as nodes_star:
        nodes.writeStar(nodes_star, tableName="output_nodes")

    end = time.time()
    diff = end - start
    print("Job duration = %dh %dmin %dsec \n" % (diff//3600, diff//60 % 60, diff % 60))


def main():
    """Change to the job working directory, then call run_job()"""
    help = """
External job for calling cryolo fine-tune training within Relion 3.1.0. Run it in the main Relion project directory, e.g.:
    external_job_cryolo_train.py --o External/cryolo_training --in_parts Select/job004/particles.star --box_size 128
"""
    parser = argparse.ArgumentParser(usage=help)
    parser.add_argument("--in_parts", help="Input particles STAR file")
    parser.add_argument("--o", dest="out_dir", help="Output directory name")
    parser.add_argument("--box_size", help="Box size (px)", type=int)
    parser.add_argument("--model", help="Cryolo training model (if not specified general is used)")
    parser.add_argument("--gpu", help='GPUs to use (e.g. "0 1 2 3")', default="0")
    parser.add_argument("--j", dest="threads", help="Not used here. required by relion")
    parser.add_argument("--pipeline_control", help="Not used here. Required by relion")

    args = parser.parse_args()

    if args.in_parts is None or args.out_dir is None or args.box_size is None:
        print("Error: --in_parts,  --o and --box_size are required params!")
        exit(1)

    if not args.in_parts.endswith(".star"):
        print("Error: --in_parts must point to a particles star file")
        exit(1)

    project_dir = os.getcwd()
    os.makedirs(args.out_dir, exist_ok=True)
    os.chdir(args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        run_job(project_dir, args)
    except:
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


if __name__ == "__main__":
    main()




# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

import numpy as np

# See https://www.awaresystems.be/imaging/tiff/tifftags/baseline.html
# BigTIFF dtypes:
# 1: BYTE 1-byte uint
# 2: ASCII 8-byte
# 3: SHORT 2-byte uint
# 5: RATIONAL 2x 4-byte uint
# 16: LONG8 8-byte uint

bigtiff_tags = {
    # tag: (name, dtype)
    256: ("ImageWidth", 3),  # nx
    257: ("ImageLength", 3),  # ny
    258: ("BitsPerSample", 3),  # data type
    259: ("Compression", 3),
    262: ("PhotometricInterpretation", 3),  # The color space of the image data.
    270: ("ImageDescription", 2),
    273: ("StripOffsets", 16),
    277: ("SamplesPerPixel", 3),
    278: ("RowsPerStrip", 3),
    279: ("StripByteCounts", 16),
    282: ("XResolution", 5),  # The number of pixels per ResolutionUnit in the ImageWidth direction.
    283: ("YResolution", 5),  # The number of pixels per ResolutionUnit in the ImageLength direction.
    284: ("PlanarConfiguration", 3),
    296: ("ResolutionUnit", 3),  # The unit of measurement for XResolution and YResolution.
    306: ("DateTime", 2),  # YYYY:MM:DD HH:MM:SS
    339: ("SampleFormat", 3),
    340: ("SMinSampleValue", 1),  # min
    341: ("SMaxSampleValue", 1),  # max
}

# See https://www.ccpem.ac.uk/mrc_format/mrc2014.php
# The code below is taken from mrcfile/dtypes.py and modified,
# is distributed under a 3-Clause BSD licence
# Copyright (c) 2016, Science and Technology Facilities Council
mrc_tags = np.dtype([
    ('NX', 'i4'),
    ('NY', 'i4'),
    ('NZ', 'i4'),
    ('MODE', 'i4'),
    ('NXSTART', 'i4'),
    ('NYSTART', 'i4'),
    ('NZSTART', 'i4'),
    ('MX', 'i4'),
    ('MY', 'i4'),
    ('MZ', 'i4'),
    ('CELLA', [  # cell dimensions in Angstroms
        ('X', 'f4'),
        ('Y', 'f4'),
        ('Z', 'f4')
    ])
])

# See EPU user manual for FEI1 MRC header details
# For Timestamp definition see:
# https://docs.microsoft.com/en-us/cpp/atl-mfc-shared/reference/coledatetime-class?redirectedfrom=MSDN&view=vs-2019
fei_tags = np.dtype([
    ('Metadata size', 'i4'),  # bytes
    ('Metadata version', 'i4'),
    ('Bitmask 1', 'u4'),
    ('Timestamp', 'f8'),  # days from December 30, 1899, at midnight
    ('Microscope type', 'S16'),
    ('D-Number', 'S16'),
    ('Application', 'S16'),
    ('Application version', 'S16'),
    ('HT', 'f8'),  # Volt
    ('Dose', 'f8'),  # electrons/m^2
    ('Alpha tilt', 'f8'),  # deg.
    ('Beta tilt', 'f8'),  # deg.
    ('X-Stage', 'f8'),  # m
    ('Y-Stage', 'f8'),  # m
    ('Z-Stage', 'f8'),  # m
    ('Tilt axis angle', 'f8'),  # deg.
    ('Dual axis rotation', 'f8'),  # deg.
    ('Pixel size X', 'f8'),  # m
    ('Pixel size Y', 'f8'),  # m
    ('Unused space', 'S48'),
    ('Defocus', 'f8'),  # m
    ('STEM Defocus', 'f8'),  # m
    ('Applied defocus', 'f8'),  # m
    ('Instrument mode', 'i4'),
    ('Projection mode', 'i4'),
    ('Objective lens mode', 'S16'),
    ('High magnification mode', 'S16'),
    ('Probe mode', 'i4'),
    ('EFTEM On', '?'),
    ('Magnification', 'f8'),
    ('Bitmask 2', 'u4'),
    ('Camera length', 'f8'),  # m
    ('Spot index', 'i4'),
    ('Illuminated area', 'f8'),  # m
    ('Intensity', 'f8'),
    ('Convergence angle', 'f8'),  # degr.
    ('Illumination mode', 'S16'),
    ('Wide convergence angle range', '?'),
    ('Slit inserted', '?'),
    ('Slit width', 'f8'),  # eV
    ('Acceleration voltage offset', 'f8'),  # Volt
    ('Drift tube voltage', 'f8'),  # Volt
    ('Energy shift', 'f8'),  # eV
    ('Shift offset X', 'f8'),  # a.u.
    ('Shift offset Y', 'f8'),  # a.u.
    ('Shift X', 'f8'),  # a.u.
    ('Shift Y', 'f8'),  # a.u.
    ('Integration time', 'f8'),  # s
    ('Binning Width', 'i4'),
    ('Binning Height', 'i4'),
    ('Camera name', 'S16'),
    ('Readout area left', 'i4'),
    ('Readout area top', 'i4'),
    ('Readout area right', 'i4'),
    ('Readout area bottom', 'i4'),
    ('Ceta noise reduction', '?'),
    ('Ceta frames summed', 'i4'),
    ('Direct detector electron counting', '?'),
    ('Direct detector align frames', '?'),
    ('Camera param reserved 0', 'i4'),
    ('Camera param reserved 1', 'i4'),
    ('Camera param reserved 2', 'i4'),
    ('Camera param reserved 3', 'i4'),
    ('Bitmask 3', 'u4'),
    ('Camera param reserved 4', 'i4'),
    ('Camera param reserved 5', 'i4'),
    ('Camera param reserved 6', 'i4'),
    ('Camera param reserved 7', 'i4'),
    ('Camera param reserved 8', 'i4'),
    ('Camera param reserved 9', 'i4'),
    ('Phase Plate', '?'),
    ('STEM Detector name', 'S16'),
    ('Gain', 'f8'),
    ('Offset', 'f8'),
    ('STEM param reserved 0', 'i4'),
    ('STEM param reserved 1', 'i4'),
    ('STEM param reserved 2', 'i4'),
    ('STEM param reserved 3', 'i4'),
    ('STEM param reserved 4', 'i4'),
    ('Dwell time', 'f8'),  # s
    ('Frame time', 'f8'),  # s
    ('Scan size left', 'i4'),
    ('Scan size top', 'i4'),
    ('Scan size right', 'i4'),
    ('Scan size bottom', 'i4'),
    ('Full scan FOV X', 'f8'),  # m
    ('Full scan FOV Y', 'f8'),  # m
    ('Element', 'S16'),
    ('Energy interval lower', 'f8'),
    ('Energy interval higher', 'f8'),
    ('Method', 'i4'),
    ('Is dose fraction', '?'),
    ('Fraction number', 'i4'),
    ('Start frame', 'i4'),
    ('End frame', 'i4'),
    ('Input stack filename', 'S80'),
    ('Bitmask 4', 'u4'),
    ('Alpha tilt min', 'f8'),  # deg.
    ('Alpha tilt max', 'f8')  # deg.
])



#!/usr/bin/env python3

""" Based on https://github.com/diamondlightsource/python-relion-yolo-it by
Sjors H.W. Scheres, Takanori Nakane, Colin M. Palmer, Donovan Webb"""

import os
import time
import argparse
from emtable import Table  # requires pip install emtable


RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"
DEBUG = 1


def run_job(project_dir, args):
    start = time.time()
    in_parts = args.in_parts
    job_dir = args.out_dir

    getPath = lambda *arglist: os.path.join(project_dir, *arglist)

    # Reading the model star file from relion
    model = in_parts.replace("_data.star", "_model.star")
    clstable = Table(fileName=getPath(model), tableName='model_classes')

    # Find classes with >= 5% particles
    good_cls = []
    for row in clstable:
        if row.rlnClassDistribution >= 0.05:
            good_cls.append(int(row.rlnReferenceImage.split("@")[0]))
    if DEBUG:
        print("Classes selected: ", good_cls)

    if len(good_cls) == 0:
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        exit(1)

    optics = Table(fileName=getPath(in_parts), tableName='optics')
    ptcls = Table(fileName=getPath(in_parts), tableName='particles')
    cols = ptcls.getColumnNames()
    out_ptcls = Table(columns=cols)
    
    for row in ptcls:
        if row.rlnClassNumber in good_cls:
            out_ptcls.addRow(*row)

    if DEBUG:
        print("Input particles: %d\nOutput particles: %d" % 
              (len(ptcls), len(out_ptcls)))

    out_star = getPath(job_dir, "particles_for_training.star")
    with open(out_star, "w") as f:
        optics.writeStar(f, tableName="optics")
        out_ptcls.writeStar(f, tableName="particles")

    end = time.time()
    diff = end - start
    print("Job duration = %dh %dmin %dsec \n" % (diff//3600, diff//60 % 60, diff % 60))


def main():
    """Change to the job working directory, then call run_job()"""
    help = """
External job to parse *_data.star from Class2D job and save particles with rlnClassDistribution >=0.05:
    external_job_sortcl2d.py --o External/best_cl2d --in_parts Class2D/job004/run_it020_data.star
"""
    parser = argparse.ArgumentParser(usage=help)
    parser.add_argument("--in_parts", help="Input *_data.star STAR file from Class2D job")
    parser.add_argument("--o", dest="out_dir", help="Output directory name")
    parser.add_argument("--j", dest="threads", help="Not used here. Required by relion")
    parser.add_argument("--pipeline_control", help="Not used here. Required by relion")

    args = parser.parse_args()

    if args.in_parts is None or args.out_dir is None:
        print("Error: --in_parts and --o are required params!")
        exit(1)

    if not args.in_parts.endswith("_data.star"):
        print("Error: --in_parts must point to a *_data.star file")
        exit(1)

    project_dir = os.getcwd()
    os.makedirs(args.out_dir, exist_ok=True)
    os.chdir(args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        run_job(project_dir, args)
    except:
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


if __name__ == "__main__":
    main()




#!/usr/bin/env python3

""" Based on https://github.com/DiamondLightSource/python-relion-yolo-it by
Sjors H.W. Scheres, Takanori Nakane, Colin M. Palmer, Donovan Webb"""

import argparse
import os
import time
import subprocess
from emtable import Table


RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"
CONDA_ENV = ". /home/gsharov/rc/conda.rc && conda activate cinderella-0.7.0"
CINDERELLA_PREDICT = "sp_cinderella_predict.py"
CINDERELLA_GEN_MODEL = "/home/gsharov/soft/cryolo/relion_lmb_cl2d_model_based_on_cinderella.h5"
DEBUG = 1


def run_job(project_dir, args):
    start = time.time()
    in_parts = args.in_parts
    job_dir = args.out_dir
    thresh = args.threshold
    model = args.model
    gpus = args.gpu

    getPath = lambda *arglist: os.path.join(project_dir, *arglist)

    if model == "None":
        model = CINDERELLA_GEN_MODEL
    else:
        model = getPath(model)

    # Reading the model star file from relion
    modelstar = in_parts.replace("_data.star", "_model.star")
    refstable = Table(fileName=getPath(modelstar), tableName='model_classes')
    refstack = refstable[0].rlnReferenceImage.split("@")[-1]

    if DEBUG:
        print("Found input class averages stack: %s" % refstack)

    # Launching cinderella
    args_dict = {
        '-i': getPath(refstack),
        '-o': 'output',
        '-w': model,
        '--gpu': gpus,
        '-t': thresh,
    }
    cmd = "%s && %s " % (CONDA_ENV, CINDERELLA_PREDICT)
    cmd += " ".join(['%s %s' % (k, v) for k, v in args_dict.items()])

    print("Running command:\n{}".format(cmd))
    proc = subprocess.Popen(cmd, shell=True)
    proc.communicate()

    if proc.returncode:
        raise Exception("Command failed with return code %d" % proc.returncode)

    # Parse output to get good classes IDs
    outfn = os.path.basename(refstack.replace(".mrcs", "_index_confidence.txt"))
    outpath = getPath(job_dir, "output", outfn)
    goodcls = []
    with open(outpath, "r") as f:
        for line in f:
            if float(line.split()[1]) > thresh:
                goodcls.append(int(line.split()[0]) + 1)
            else:
                break

    if DEBUG:
        print("Parsing output file: %s\nGood classes: %s" % (outpath, goodcls))

    if len(goodcls) == 0:
        print("No good classes found. Job stopped.")
        end = time.time()
        diff = end - start
        print("Job duration = %dh %dmin %dsec \n" % (diff // 3600, diff // 60 % 60, diff % 60))
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        exit(1)

    # Create output star file for Relion to use
    optics = Table(fileName=getPath(in_parts), tableName='optics')
    ptcls = Table(fileName=getPath(in_parts), tableName='particles')
    cols = ptcls.getColumnNames()
    out_ptcls = Table(columns=cols)
    
    for row in ptcls:
        if row.rlnClassNumber in goodcls:
            out_ptcls.addRow(*row)

    if DEBUG:
        print("Input particles: %d\nOutput particles: %d" % 
              (len(ptcls), len(out_ptcls)))

    out_star = getPath(job_dir, "particles_for_training.star")
    with open(out_star, "w") as f:
        optics.writeStar(f, tableName="optics")
        out_ptcls.writeStar(f, tableName="particles")

    end = time.time()
    diff = end - start
    print("Job duration = %dh %dmin %dsec \n" % (diff//3600, diff//60 % 60, diff % 60))


def main():
    """Change to the job working directory, then call run_job()"""
    help = """
External job for calling cinderella within Relion 3.1.0. Run it in the main Relion project directory, e.g.:
    external_job_cinderella.py --o External/cinderella_bestclasses2d --in_parts Class2D/job004/run_it025_data.star --threshold 0.7
"""
    parser = argparse.ArgumentParser(usage=help)
    parser.add_argument("--in_parts", help="Input data STAR file from Class2D job")
    parser.add_argument("--o", dest="out_dir", help="Output directory name")
    parser.add_argument("--threshold", help="Confidence threshold (default = 0.7)", type=float, default=0.7)
    parser.add_argument("--model", help="Cinderella prediction model (if not specified general is used)", default="None")
    parser.add_argument("--gpu", help='GPUs to use (e.g. "0 1 2 3")', default="0")
    parser.add_argument("--pipeline_control", help="Not used here. Required by relion")
    parser.add_argument("--j", dest="threads", help="Not used here. Required by relion")

    args = parser.parse_args()

    if args.in_parts is None or args.out_dir is None:
        print("Error: --in_parts and --o are required params!")
        exit(1)

    if not args.in_parts.endswith("_data.star"):
        print("Error: --in_parts must point to a data STAR file")
        exit(1)

    project_dir = os.getcwd()
    os.makedirs(args.out_dir, exist_ok=True)
    os.chdir(args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        run_job(project_dir, args)
    except:
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


if __name__ == "__main__":
    main()



# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

# use polling as required for watchdog over NFS/CIFS
from watchdog.observers.polling import PollingObserver as Observer
from watchdog.events import RegexMatchingEventHandler

from .config import *
from .utils.misc import getUsername, setParticleSizes
from .parser import Parser
from .schedule import setupRelion, setupScipion


class WatchDog:
    def __init__(self):
        self.observer = Observer()

    def start_daemon(self, path):
        """ Watch for new xml/mdoc files in METADATA_PATH. """
        if DEF_SOFTWARE == "EPU":
            regex = r".*/%s.*/Images-Disc\d/GridSquare_.*/Data/FoilHole_.*_Data_.*.xml$" % DEF_PREFIX
        else:
            regex = r".*/%s.*/.*.tif.mdoc$" % DEF_PREFIX
        event_handler = RegexMatchingEventHandler(regexes=[regex],
                                                  ignore_regexes=[],
                                                  ignore_directories=True,
                                                  case_sensitive=True)
        event_handler.on_created = self.on_created

        if DEBUG:
            print("Watchdog started for: %s\n\tregex: %s" %
                  (path, regex))

        self.observer.schedule(event_handler, path, recursive=True)
        self.observer.start()
        self.observer.join()

    def on_created(self, event):
        mdFn = event.src_path
        if DEBUG:
            print("File creation detected: %s" % mdFn)
        self.observer.stop()
        start_app(mdFn)


def start_app(mdFn):
    """ Simulate GUI mode, setup acqDict and run parsers. """
    if DEF_SOFTWARE == "EPU":
        mdPath = "/".join(mdFn.split("/")[:-4])
    else:
        mdPath = "/".join(mdFn.split("/")[:-1])
    model = Parser()
    model.setSoftware(DEF_SOFTWARE)
    model.setPipeline(DEF_PIPELINE)
    model.setMdPath(mdPath)
    model.setFn(mdFn)

    username, uid = getUsername(mdPath)
    model.setUser(username, uid)
    model.acqDict['User'] = model.getUser()

    if DEBUG:
        print("\n\nInput params: ",
              model.getSoftware(),
              model.getMdPath(),
              model.getUser(),
              model.getPipeline())
        print("\nFiles found: %s\n" % mdFn) if DEBUG else ""

    if DEF_SOFTWARE == 'EPU':
        model.parseImgEpu(mdFn)
    else:  # SerialEM
        model.parseImgMdoc(mdFn)

    model.calcDose()
    model.guessDataDir(wait=True)
    model.acqDict['Picker'] = DEF_PICKER
    setParticleSizes(model)

    if DEBUG:
        print("\nFinal parameters:\n")
        for k, v in sorted(model.acqDict.items()):
            print(k, v)
        print('\n')

    if DEF_PIPELINE == 'Relion':
        setupRelion(model.acqDict)
    else:
        setupScipion(model.acqDict)



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

import os
import sys
from PyQt5.QtWidgets import (QGridLayout, QLabel, QMessageBox,
                             QHBoxLayout, QVBoxLayout, QRadioButton,
                             QPushButton, QWizard, QGroupBox,
                             QSizePolicy, QLineEdit, QFileDialog,
                             QCheckBox, QApplication, QWizardPage,
                             QButtonGroup, QSpinBox)
from PyQt5.QtCore import Qt

from .config import *
from .utils.misc import getUsername
from .parser import Parser
from .schedule import setupRelion, setupScipion


class App(QWizard):
    """ Main class that runs the GUI. """
    model = Parser()

    def __init__(self, parent=None):
        super(App, self).__init__(parent, flags=Qt.WindowFlags())
        self.title = 'MDCatch v%s - metadata parser' % VERSION
        self.width = 640
        self.height = 320
        self.initUI()

    def initUI(self):
        self.page1 = Page1()
        self.addPage(self.page1)
        self.page2 = Page2()
        self.addPage(self.page2)
        self.button(QWizard.BackButton).clicked.connect(self.page1.reset)
        self.button(QWizard.FinishButton).clicked.connect(self.page2.onFinish)
        self.setWindowTitle(self.title)
        # remove Help button from the window
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self.resize(self.width, self.height)

    @staticmethod
    def showDialog(title, text, mtype='error', extra=None):
        """ Dialog message with a warning or an error. """
        msg = QMessageBox()
        if mtype == 'error':
            msg.setIcon(QMessageBox.Critical)
        else:
            msg.setIcon(QMessageBox.Information)
        msg.setWindowTitle(title)
        msg.setText(text)
        if extra is not None:
            msg.setDetailedText(extra)
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec_()


class Page1(QWizardPage):
    """ Page 1 with user input params. """
    def __init__(self, parent=None):
        super(Page1, self).__init__(parent)
        self.setSubTitle("Input parameters")
        self.mainLayout = QGridLayout()
        self.mainLayout.addLayout(self.group1(), 0, 0)
        self.mainLayout.addLayout(self.group2(), 0, 1)
        self.setLayout(self.mainLayout)

    def group1(self):
        """ Label widgets at row 0, col 0. """
        vbox = QVBoxLayout()
        label_soft = QLabel('Software')
        label_path = QLabel('Path')
        label_pipeline = QLabel('Run pipeline in')
        label_picker = QLabel('Particle picker')
        label_diam = QLabel('Particle diameter (A)')
        label_diam.setToolTip(help_picker)

        vbox.addWidget(label_soft, alignment=Qt.Alignment())
        vbox.addWidget(label_path, alignment=Qt.Alignment())
        vbox.addWidget(label_pipeline, alignment=Qt.Alignment())
        vbox.addWidget(label_picker, alignment=Qt.Alignment())
        vbox.addWidget(label_diam, alignment=Qt.Alignment())

        return vbox

    def group2(self):
        """ Actual input widgets at row 0, col 1. """
        grid = QVBoxLayout()

        # software type
        hbox_soft = QHBoxLayout()
        hbox_soft.setAlignment(Qt.AlignLeft)
        btgroup_soft = QButtonGroup()

        button_epu = self.addRadioButton("EPU", default=DEF_SOFTWARE == "EPU")
        button_sem = self.addRadioButton("SerialEM", default=DEF_SOFTWARE == "SerialEM")

        btgroup_soft.addButton(button_epu)
        btgroup_soft.addButton(button_sem)
        btgroup_soft.buttonClicked.connect(lambda: self.updSoftware(btgroup_soft))
        hbox_soft.addWidget(button_epu, alignment=Qt.Alignment())
        hbox_soft.addWidget(button_sem, alignment=Qt.Alignment())
        grid.addLayout(hbox_soft)

        # path box
        hbox_path = QHBoxLayout()
        self.rawPath = QLineEdit()
        self.rawPath.setMinimumWidth(300)
        self.rawPath.setReadOnly(True)
        self.rawPath.setText(METADATA_PATH)
        self.rawPath.setToolTip(help_message)

        button_browse = QPushButton('Browse')
        button_browse.setToolTip(help_message)
        button_browse.clicked.connect(lambda: self.browseSlot(self.rawPath))

        hbox_path.addWidget(self.rawPath, alignment=Qt.Alignment())
        hbox_path.addWidget(button_browse, alignment=Qt.Alignment())
        grid.addLayout(hbox_path)

        # pipeline
        hbox_pipeline = QHBoxLayout()
        hbox_pipeline.setAlignment(Qt.AlignLeft)
        btgroup_pipeline = QButtonGroup()

        button_relion = self.addRadioButton("Relion", default=DEF_PIPELINE == "Relion")
        button_scipion = self.addRadioButton("Scipion", default=DEF_PIPELINE == "Scipion")

        btgroup_pipeline.addButton(button_relion)
        btgroup_pipeline.addButton(button_scipion)
        btgroup_pipeline.buttonClicked.connect(lambda: self.updPipeline(btgroup_pipeline))
        hbox_pipeline.addWidget(button_relion, alignment=Qt.Alignment())
        hbox_pipeline.addWidget(button_scipion, alignment=Qt.Alignment())
        grid.addLayout(hbox_pipeline)

        # particle picker
        hbox_picker = QHBoxLayout()
        hbox_picker.setAlignment(Qt.AlignLeft)
        btgroup_picker = QButtonGroup()

        self.button_cryolo = self.addRadioButton("crYOLO", default=DEF_PICKER == "crYOLO")
        self.button_topaz = self.addRadioButton("Topaz", default=DEF_PICKER == "Topaz")
        self.button_logpick = self.addRadioButton("LogPicker", default=DEF_PICKER == "LogPicker")

        btgroup_picker.addButton(self.button_cryolo)
        btgroup_picker.addButton(self.button_topaz)
        btgroup_picker.addButton(self.button_logpick)
        btgroup_picker.buttonClicked.connect(lambda: self.updPicker(btgroup_picker))
        hbox_picker.addWidget(self.button_cryolo, alignment=Qt.Alignment())
        hbox_picker.addWidget(self.button_topaz, alignment=Qt.Alignment())
        hbox_picker.addWidget(self.button_logpick, alignment=Qt.Alignment())
        grid.addLayout(hbox_picker)

        # size box
        hbox_diam = QHBoxLayout()
        self.label_diamMin = QLabel('')
        self.label_diamMin.setToolTip(help_picker)
        hbox_diam.addWidget(self.label_diamMin)

        self.spbox_diamMin = QSpinBox()
        self.spbox_diamMin.setToolTip(help_picker)
        self.spbox_diamMin.setRange(0, 9999)
        self.spbox_diamMin.setValue(LOGPICKER_SIZES[0] if DEF_PICKER != "crYOLO" else 0)
        self.spbox_diamMin.setFixedSize(60, 25)
        hbox_diam.addWidget(self.spbox_diamMin)

        self.label_diamMax = QLabel('max')
        self.label_diamMax.setToolTip(help_picker)
        self.label_diamMax.setVisible(self.button_logpick.isChecked())
        hbox_diam.addWidget(self.label_diamMax)

        self.spbox_diamMax = QSpinBox()
        self.spbox_diamMax.setToolTip(help_picker)
        self.spbox_diamMax.setVisible(self.button_logpick.isChecked())
        self.spbox_diamMax.setRange(10, 9999)
        self.spbox_diamMax.setValue(LOGPICKER_SIZES[1])
        self.spbox_diamMax.setFixedSize(60, 25)
        hbox_diam.addWidget(self.spbox_diamMax)
        hbox_diam.setAlignment(Qt.AlignLeft)
        grid.addLayout(hbox_diam)

        return grid

    def updSoftware(self, btgroup):
        bt = btgroup.checkedButton()
        App.model.setSoftware(bt.text())

    def updPipeline(self, btgroup):
        bt = btgroup.checkedButton()
        App.model.setPipeline(bt.text())

    def updPicker(self, btgroup):
        bt = btgroup.checkedButton()
        if self.button_cryolo.isChecked():  # crYOLO
            self.label_diamMin.setText("")
            self.spbox_diamMin.setValue(0)
            self.label_diamMax.setVisible(False)
            self.spbox_diamMax.setVisible(False)
        elif self.button_topaz.isChecked():  # Topaz
            self.label_diamMin.setText("")
            self.spbox_diamMin.setValue(LOGPICKER_SIZES[0])
            self.label_diamMax.setVisible(False)
            self.spbox_diamMax.setVisible(False)
        elif self.button_logpick.isChecked():  # LogPicker
            self.label_diamMin.setText("min")
            self.spbox_diamMin.setValue(LOGPICKER_SIZES[0])
            self.label_diamMax.setVisible(True)
            self.spbox_diamMax.setVisible(True)

        App.model.setPicker(bt.text())
        App.model.setSize(self.spbox_diamMin.value(), self.spbox_diamMax.value())

    def browseSlot(self, var):
        """ Called when "Browse" is pressed. """
        folder = METADATA_PATH if var.text() is None else var.text()
        path = QFileDialog.getExistingDirectory(self, "Select Directory",
                                                folder,
                                                options=QFileDialog.ShowDirsOnly)
        if path:
            self.refreshPath(path)

    def refreshPath(self, path):
        """ Update line widget with selected path. """
        App.model.setMdPath(path)
        self.rawPath.setText(App.model.getMdPath())

    def validatePage(self):
        """ Executed when Next is pressed.
        Returns True or False. """
        App.model.setSize(self.spbox_diamMin.value(), self.spbox_diamMax.value())

        if App.model.getMdPath() is None:
            App.model.setMdPath(METADATA_PATH)

        username, uid = getUsername(App.model.getMdPath())
        App.model.setUser(username, uid)

        if DEBUG:
            print("\n\nInput params: ",
                  [App.model.getSoftware(),
                   App.model.getMdPath(),
                   App.model.getUser(),
                   App.model.getPipeline(),
                   App.model.getPicker(),
                   App.model.getSize()])

        prog = App.model.getSoftware()
        fnList = App.model.guessFn(prog)

        if fnList is None:
            App.showDialog("ERROR", "No matching files found!\n\n" + help_message)
            return False
        else:
            print("\nFiles found: %s\n" % fnList) if DEBUG else ""
            App.model.setFn(fnList)
            return True

    def reset(self):
        """ Executed when "Back" is pressed. """
        App.model.acqDict.clear()
        App.model.__init__()
        # keep the old path until updated
        App.model.setMdPath(self.rawPath.text())

    def addRadioButton(self, choice, default=False):
        """ Util func to add QRadioButton widget. """
        rb = QRadioButton(choice)
        if default:
            rb.setChecked(True)

        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        rb.setSizePolicy(sizePolicy)
        return rb


class Page2(QWizardPage):
    """ Page 2 with parsed results. """
    def __init__(self, parent=None):
        super(Page2, self).__init__(parent)
        self.mainLayout = QGridLayout()
        self.mainLayout.addWidget(self.group1(), 0, 0)
        self.mainLayout.addWidget(self.group2(), 0, 1)
        self.doCalcBox = False
        self.setLayout(self.mainLayout)

    def initializePage(self):
        # executed before showing page 2
        acqDict = App.model.acqDict
        prog = App.model.getSoftware()
        fnList = App.model.getFn()

        if prog == 'EPU':
            App.model.parseImgEpu(fnList)
        else:  # SerialEM
            App.model.parseImgMdoc(fnList)

        self.addPtclSizeWidgets(acqDict)
        App.model.calcDose()
        App.model.guessDataDir()

        self.setSubTitle("Found the following metadata from %s session:" % prog)

        scopeID = acqDict['MicroscopeID']
        time = round(float(acqDict['ExposureTime']), 3)
        dosepf = round(float(acqDict['DosePerFrame']), 2)
        px = round(float(acqDict['PixelSpacing']), 4)

        self.scope_name.setText(SCOPE_DICT[scopeID][0])
        self.kv.setText(acqDict['Voltage'])
        self.cs.setText(acqDict['Cs'])
        self.px.setText(str(px))

        vpp = acqDict['PhasePlateUsed']
        if vpp in ['true', 'True', True]:
            self.vpp.setChecked(True)
        else:
            self.vpp.setChecked(False)

        self.camera_name.setText(acqDict['Detector'])
        self.mode.setText(acqDict['Mode'])
        self.time.setText(str(time))
        self.frames.setText(acqDict['NumSubFrames'])
        self.dosepf.setText(str(dosepf))
        self.gain.setText(os.path.basename(acqDict['GainReference']))
        self.defects.setText(os.path.basename(acqDict['DefectFile']))

    def group1(self):
        """ Widgets at row 0, col 0. """
        groupBox = QGroupBox("Microscope")
        name = QLabel("Name")
        kv = QLabel("Voltage (kV)")
        cs = QLabel("Cs (mm)")
        vpp = QLabel("Phase plate")
        px = QLabel("Pixel size (A)")

        self.scope_name = QLabel()
        self.kv = QLabel()
        self.cs = QLabel()
        self.vpp = QCheckBox()
        self.px = self.addLine(50, 5, Qt.AlignRight)

        vbox = QGridLayout()
        for num, i in enumerate([name, kv, cs, px, vpp]):
            vbox.addWidget(i, num, 0)

        for num, i in enumerate([self.scope_name, self.kv,
                                 self.cs, self.px, self.vpp]):
            vbox.addWidget(i, num, 1)

        groupBox.setLayout(vbox)
        return groupBox

    def group2(self):
        """ Widgets at row 0, col 1. """
        groupBox = QGroupBox("Detector")
        name = QLabel("Name")
        mode = QLabel("Mode")
        time = QLabel("Exposure time (s)")
        frames = QLabel("Frames")
        dosepf = QLabel("Dose per frame (e/A²)")
        gain = QLabel("Gain reference")
        defects = QLabel("Defects file")

        self.camera_name = QLabel()
        self.mode = QLabel()
        self.time = QLabel()
        self.frames = QLabel()
        self.gain = QLabel()
        self.defects = QLabel()
        self.dosepf = self.addLine(50, 5, Qt.AlignRight)

        vbox = QGridLayout()
        for num, i in enumerate([name, mode, time, frames,
                                 dosepf, gain, defects]):
            vbox.addWidget(i, num, 0)

        for num, i in enumerate([self.camera_name, self.mode,
                                 self.time, self.frames,
                                 self.dosepf, self.gain,
                                 self.defects]):
            vbox.addWidget(i, num, 1)

        groupBox.setLayout(vbox)
        return groupBox

    def group3(self):
        groupBox = QGroupBox("Recommended options")

        box = QLabel("Box size (px)")
        mask = QLabel("Mask size (px)")
        box2 = QLabel("Downscale to (px)")

        self.box = self.addLine(50, 4, Qt.AlignRight)
        self.mask = self.addLine(50, 4, Qt.AlignRight)
        self.box_bin = self.addLine(50, 4, Qt.AlignRight)

        vbox = QGridLayout()
        for num, i in enumerate([box, mask, box2]):
            vbox.addWidget(i, num, 0)

        for num, i in enumerate([self.box, self.mask,
                                 self.box_bin]):
            vbox.addWidget(i, num, 1)

        groupBox.setLayout(vbox)

        return groupBox

    def addPtclSizeWidgets(self, acqDict):
        picker = App.model.getPicker()
        sizes = App.model.getSize()
        acqDict['PtclSizes'] = sizes
        acqDict['Picker'] = picker

        if picker == 'crYOLO' and sizes[0] == 0:
            pass  # automated estimation, no params needed
        else:
            self.doCalcBox = True
            App.model.calcBox(picker)
            self.mainLayout.addWidget(self.group3(), 1, 0)
            self.box.setText(acqDict['BoxSize'])
            self.mask.setText(acqDict['MaskSize'])
            self.box_bin.setText(acqDict['BoxSizeSmall'])

    def onFinish(self):
        """ Finish is pressed, we need to update all editable vars. """
        App.model.acqDict['User'] = App.model.getUser()
        App.model.acqDict['DosePerFrame'] = self.dosepf.text()
        App.model.acqDict['PixelSpacing'] = self.px.text()
        App.model.acqDict['PhasePlateUsed'] = self.vpp.isChecked()

        if self.doCalcBox:
            App.model.acqDict['BoxSize'] = self.box.text()
            App.model.acqDict['MaskSize'] = self.mask.text()
            App.model.acqDict['BoxSizeSmall'] = self.box_bin.text()

        print("\nFinal parameters:\n")
        for k, v in sorted(App.model.acqDict.items()):
            print(k, v)
        print('\n')

        if App.model.getPipeline() == 'Relion':
            setupRelion(App.model.acqDict)
        else:
            setupScipion(App.model.acqDict)

    def addLine(self, size, length, align):
        """ Util func to add LineEdit widget. """
        line = QLineEdit()
        line.setMinimumWidth(size)
        line.setMaximumWidth(size)
        line.setMaxLength(length)
        line.setAlignment(align)
        return line


def main():
    args = sys.argv
    help = "Usage: mdcatch [--watch]\nBy default starts a GUI, "\
           "use '--watch' for daemon mode. It will watch METADATA_PATH folder."
    if len(args) > 1:
        if args[1] in ['-h', '--help']:
            print(help)
        elif args[1] == '--watch':
            from .watcher import WatchDog
            watch = WatchDog()
            watch.start_daemon(METADATA_PATH)
        else:
            print("Unrecognized arguments.\n%s" % help)
    else:
        app = QApplication(sys.argv)
        app.setStyle('Fusion')
        wizard = App()
        wizard.show()
        sys.exit(app.exec_())



# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
from mdcatch.config import VERSION

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='MDCatch',
    version=VERSION,
    description='Fetch metadata from EPU / SerialEM and launch on-the-fly pre-processing',
    long_description=long_description,
    url='https://github.com/azazellochg/MDCatch',
    author='Grigory Sharov',
    author_email='gsharov@mrc-lmb.cam.ac.uk',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3'
    ],
    keywords='cryo-em python qt5 epu serialem',
    packages=find_packages(),
    install_requires=['pyqt5', 'numpy', 'emtable', 'watchdog'],
    entry_points={
        'console_scripts': [
            'mdcatch=mdcatch.run:main'
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/azazellochg/MDCatch/issues',
        'Source': 'https://github.com/azazellochg/MDCatch',
    },
)



#!/usr/bin/env python3

""" Based on https://github.com/DiamondLightSource/python-relion-yolo-it by
Sjors H.W. Scheres, Takanori Nakane, Colin M. Palmer, Donovan Webb"""

import argparse
import json
import os
import shutil
import time
from glob import glob
import subprocess
import math
from emtable import Table  # requires pip install emtable


RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"
DONE_MICS = "done_mics.txt"
CONDA_ENV = ". /home/gsharov/rc/conda.rc && conda activate cryolo-1.7.5"
CRYOLO_PREDICT = "cryolo_predict.py"
CRYOLO_GEN_MODEL = "/home/gsharov/soft/cryolo/gmodel_phosnet_202005_N63_c17.h5"
CRYOLO_JANNI_MODEL = "/home/gsharov/soft/cryolo/gmodel_janni_20190703.h5"
DEBUG = 1


def run_job(project_dir, args):
    start = time.time()
    in_mics = args.in_mics
    job_dir = args.out_dir
    thresh = args.threshold
    box_size = args.box_size
    distance = 0
    model = args.model
    gpus = args.gpu
    threads = args.threads

    getPath = lambda *arglist: os.path.join(project_dir, *arglist)

    if model == "None":
        model = CRYOLO_GEN_MODEL
    else:
        model = getPath(model)

    # Making a cryolo config file
    json_dict = {"model": {
        "architecture": "PhosaurusNet",
        "input_size": 1024,
        "max_box_per_image": 600,
        "filter": [
            0.1,
            "filtered_tmp/"
        ]
    }}

    if box_size:  # is not 0
        json_dict["model"]["anchors"] = [int(box_size), int(box_size)]
        distance = int(box_size / 2)  # use half the box_size

    if DEBUG:
        print("Using following config.json: ", json_dict)

    with open("config.json", "w") as json_file:
        json.dump(json_dict, json_file, indent=4)

    # Reading the micrographs star file from relion
    mictable = Table(fileName=getPath(in_mics), tableName='micrographs')

    # Arranging files for cryolo: making symlinks for mics
    done_mics = []
    mic_dirs = []
    if os.path.exists(DONE_MICS):
        with open(DONE_MICS, "r") as f:
            done_mics.extend(f.read().splitlines())
    if DEBUG:
        print("Current done_mics: ", done_mics)

    mic_fns = mictable.getColumnValues("rlnMicrographName")
    input_job = "/".join(mic_fns[0].split("/")[:2])
    keys = ["/".join(i.split("/")[2:]) for i in mic_fns]  # remove JobType/jobXXX
    values = [os.path.splitext(i)[0] + "_cryolo.star" for i in keys]  # _cryolo.star
    mic_dict = {k: v for k, v in zip(keys, values)}

    for mic in mic_dict:
        if DEBUG:
            print("Processing mic: ", mic)
        mic_dir = os.path.dirname(mic)
        if not os.path.isdir(mic_dir):
            os.makedirs(mic_dir)
        if mic_dir not in mic_dirs:
            mic_dirs.append(mic_dir)
            print("Added folder %s to the mic_dirs" % mic_dir)
        if mic not in done_mics:
            inputfn = getPath(input_job, mic)
            outfn = getPath(job_dir, mic)

            # if mic had no coords picked, symlink still exists
            if os.path.exists(outfn):
                os.remove(outfn)
                continue
            else:
                os.symlink(inputfn, outfn)
                if DEBUG:
                    print("Link %s --> %s" % (inputfn, outfn))

    if len(mic_dict.keys()) == len(done_mics):
        print("All mics picked! Nothing to do.")
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()
        exit(0)

    # Launching cryolo
    args_dict = {
        '--conf': "config.json",
        '--output': 'output',
        '--weights': model,
        '--gpu': "%s" % gpus.replace('"', ''),
        '--threshold': thresh,
        '--distance': distance,
        '--cleanup': "",
        '-nc': -1  # threads
    }
    cmd = "%s && %s " % (CONDA_ENV, CRYOLO_PREDICT)
    cmd += " ".join(['%s %s' % (k, v) for k, v in args_dict.items()])
    for i in mic_dirs:
        cmd += " --input %s" % i

    print("Running command:\n{}".format(cmd))
    proc = subprocess.Popen(cmd, shell=True)
    proc.communicate()

    if proc.returncode:
        raise Exception("Command failed with return code %d" % proc.returncode)

    # Moving output star files for Relion to use
    with open(DONE_MICS, "a+") as f:
        for mic in mic_dict:
            mic_base = os.path.basename(mic)
            coord_cryolo = os.path.splitext(mic_base)[0] + ".star"
            coord_cryolo = getPath(job_dir, "output", "STAR", coord_cryolo)
            coord_relion = mic_dict[mic]
            # To ensure that cryolo doesn't pick from already done mics
            if os.path.exists(coord_cryolo):
                f.write("%s\n" % mic)
                os.rename(coord_cryolo, getPath(job_dir, coord_relion))
                os.remove(mic)  # clean up
                if DEBUG:
                    print("Moved %s to %s" % (coord_cryolo, getPath(job_dir, coord_relion)))

    # Required output mics star file
    with open("coords_suffix_cryolo.star", "w") as mics_star:
        mics_star.write(in_mics)

    # Required output nodes star file
    nodes = Table(columns=['rlnPipeLineNodeName', 'rlnPipeLineNodeType'])
    nodes.addRow(os.path.join(job_dir, "coords_suffix_cryolo.star"), "2")
    with open("RELION_OUTPUT_NODES.star", "w") as nodes_star:
        nodes.writeStar(nodes_star, tableName="output_nodes")

    outputFn = getPath(job_dir, "output_for_relion.star")
    if not os.path.exists(outputFn):
        # get estimated box size
        summaryfn = getPath(job_dir, "output/DISTR",
                            'size_distribution_summary*.txt')
        with open(glob(summaryfn)[0]) as f:
            for line in f:
                if line.startswith("MEAN,"):
                    estim_sizepx = int(line.split(",")[-1])
        print("\nFound estimated box size %d px from %s" % (estim_sizepx, f.name))
    
        # calculate diameter, original (boxSize) and downsampled (boxSizeSmall) box
        optics = Table(fileName=getPath(in_mics), tableName='optics')
        angpix = float(optics[0].rlnMicrographPixelSize)
        # use + 20% for diameter
        diam = int(estim_sizepx * angpix * 1.2)
        # use +30% for box size, make it even
        boxSize = 1.3 * estim_sizepx
        boxSize = math.ceil(boxSize / 2.) * 2
    
        # from relion_it.py script
        # Authors: Sjors H.W. Scheres, Takanori Nakane & Colin M. Palmer
        boxSizeSmall = None
        for box in (32, 48, 64, 96, 128, 160, 192, 256, 288, 300, 320,
                    360, 384, 400, 420, 450, 480, 512, 640, 768,
                    896, 1024):
            # Don't go larger than the original box
            if box > boxSize:
                boxSizeSmall = boxSize
                break
            # If Nyquist freq. is better than 8.5 A, use this
            # downscaled box, otherwise continue to next size up
            small_box_angpix = angpix * boxSize / box
            if small_box_angpix < 4.25:
                boxSizeSmall = box
                break
    
        print("\nEstimated parameters:\n\tDiameter (A): %d\n\tBox size (px): %d\n"
              "\tBox size binned (px): %d" % (diam, boxSize, boxSizeSmall))
    
        # output all params into a star file
        tableCryolo = Table(columns=['rlnParticleDiameter',
                                     'rlnOriginalImageSize',
                                     'rlnImageSize'])
        tableCryolo.addRow(diam, boxSize, boxSizeSmall)
        with open(outputFn, "w") as f:
            tableCryolo.writeStar(f, tableName='picker')

    end = time.time()
    diff = end - start
    print("Job duration = %dh %dmin %dsec \n" % (diff//3600, diff//60 % 60, diff % 60))


def main():
    """Change to the job working directory, then call run_job()"""
    help = """
External job for calling cryolo within Relion 3.1.0. Run it in the main Relion project directory, e.g.:
    external_job_cryolo.py --o External/cryolo_picking --in_mics CtfFind/job004/micrographs_ctf.star --box_size 128 --threshold 0.3 --gpu "0 1"
"""
    parser = argparse.ArgumentParser(usage=help)
    parser.add_argument("--in_mics", help="Input micrographs STAR file")
    parser.add_argument("--o", dest="out_dir", help="Output directory name")
    parser.add_argument("--j", dest="threads", help="Number of CPU threads (default = 1)", type=int, default=1)
    parser.add_argument("--box_size", help="Box size (default = 0 means it's estimated)", type=int, default=0)
    parser.add_argument("--threshold", help="Threshold for picking (default = 0.3)", type=float, default=0.3)
    parser.add_argument("--model", help="Cryolo training model (if not specified general is used)", default="None")
    parser.add_argument("--gpu", help='GPUs to use (e.g. "0 1 2 3")', default="0")
    parser.add_argument("--pipeline_control", help="Not used here. Required by relion")

    args = parser.parse_args()

    if args.in_mics is None or args.out_dir is None:
        print("Error: --in_mics and --o are required params!")
        exit(1)

    if not args.in_mics.endswith(".star"):
        print("Error: --in_mics must point to a micrographs star file")
        exit(1)

    project_dir = os.getcwd()
    os.makedirs(args.out_dir, exist_ok=True)
    os.chdir(args.out_dir)
    if os.path.isfile(RELION_JOB_FAILURE_FILENAME):
        os.remove(RELION_JOB_FAILURE_FILENAME)
    if os.path.isfile(RELION_JOB_SUCCESS_FILENAME):
        os.remove(RELION_JOB_SUCCESS_FILENAME)
    try:
        run_job(project_dir, args)
    except:
        open(RELION_JOB_FAILURE_FILENAME, "w").close()
        raise
    else:
        open(RELION_JOB_SUCCESS_FILENAME, "w").close()


if __name__ == "__main__":
    main()



# **************************************************************************
# *
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [1]
# *
# * [1] MRC Laboratory of Molecular Biology, MRC-LMB
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'gsharov@mrc-lmb.cam.ac.uk'
# *
# **************************************************************************

""" This script parses FEI EPU xml file. """

import math
import xml.etree.ElementTree as ET

from ..config import DEBUG, SCOPE_DICT


def parseXml(fn):
    acqDict = dict()
    tree = ET.parse(fn)
    root = tree.getroot()
    nspace = {'so': '{http://schemas.datacontract.org/2004/07/Fei.SharedObjects}',
              'ar': '{http://schemas.microsoft.com/2003/10/Serialization/Arrays}',
              'fr': '{http://schemas.datacontract.org/2004/07/Fei.Applications.Common.Omp.Interface}',
              'tp': '{http://schemas.datacontract.org/2004/07/Fei.Types}'}

    items = {'ExposureTime': "./{so}microscopeData/{so}acquisition/{so}camera/{so}ExposureTime",
             'Detector': "./{so}microscopeData/{so}acquisition/{so}camera/{so}Name",
             'GunLens': "./{so}microscopeData/{so}gun/{so}GunLens",
             'SpotSize': "./{so}microscopeData/{so}optics/{so}SpotIndex",
             'Magnification': "./{so}microscopeData/{so}optics/{so}TemMagnification/{so}NominalMagnification",
             'BeamSize': "./{so}microscopeData/{so}optics/{so}BeamDiameter",
             'Voltage': "./{so}microscopeData/{so}gun/{so}AccelerationVoltage",
             'MicroscopeID': "./{so}microscopeData/{so}instrument/{so}InstrumentID",
             'PixelSpacing': "./{so}SpatialScale/{so}pixelSize/{so}x/{so}numericValue",
             'EPUversion': "./{so}microscopeData/{so}core/{so}ApplicationSoftwareVersion",
             'BeamTiltX': "./{so}microscopeData/{so}optics/{so}BeamTilt/{tp}_x",
             'BeamTiltY': "./{so}microscopeData/{so}optics/{so}BeamTilt/{tp}_y",
             'BeamShiftX': "./{so}microscopeData/{so}optics/{so}BeamShift/{tp}_x",
             'BeamShiftY': "./{so}microscopeData/{so}optics/{so}BeamShift/{tp}_y",
             'ImageShiftX': "./{so}microscopeData/{so}optics/{so}ImageShift/{tp}_x",
             'ImageShiftY': "./{so}microscopeData/{so}optics/{so}ImageShift/{tp}_y",
             }

    for key in items:
        acqDict[key] = root.find(items[key].format(**nspace)).text

    if acqDict['MicroscopeID'] in SCOPE_DICT:
        acqDict['Cs'] = SCOPE_DICT[acqDict['MicroscopeID']][1]

    acqDict['BeamSize'] = float(acqDict['BeamSize']) * math.pow(10, 6)
    acqDict['Voltage'] = int(acqDict['Voltage']) // 1000

    # get cameraSpecificInput: ElectronCountingEnabled, SuperResolutionFactor etc.
    customDict = dict()
    keys = "./{so}microscopeData/{so}acquisition/{so}camera/{so}CameraSpecificInput/{ar}KeyValueOfstringanyType/{ar}Key"
    values = "./{so}microscopeData/{so}acquisition/{so}camera/{so}CameraSpecificInput/{ar}KeyValueOfstringanyType/{ar}Value"
    for k, v in zip(root.findall(keys.format(**nspace)), root.findall(values.format(**nspace))):
        customDict[k.text] = v.text

    # check if counting/super-res is enabled
    sr = 1.0
    if customDict['ElectronCountingEnabled'] == 'true':
        sr = float(customDict['SuperResolutionFactor'])  # 1 - counting, 2 - super-res
        acqDict['Mode'] = 'Counting' if sr == 1.0 else 'Super-resolution'

    acqDict['PixelSpacing'] = float(acqDict['PixelSpacing']) * math.pow(10, 10) / sr

    if acqDict['Detector'] == 'EF-CCD':
        elem = "./{so}microscopeData/{so}acquisition/{so}camera/{so}CameraSpecificInput/{ar}KeyValueOfstringanyType/{ar}Value/{fr}NumberOffractions"
        acqDict['NumSubFrames'] = root.find(elem.format(**nspace)).text
    else:
        # count number of DoseFractions for Falcon 3
        elem = "./{so}microscopeData/{so}acquisition/{so}camera/{so}CameraSpecificInput/{ar}KeyValueOfstringanyType/{ar}Value/{fr}DoseFractions"
        acqDict['NumSubFrames'] = len(root.find(elem.format(**nspace)))

    # get customData: Dose, DoseOnCamera, PhasePlateUsed, AppliedDefocus
    customDict = dict()
    keys = "./{so}CustomData/{ar}KeyValueOfstringanyType/{ar}Key"
    values = "./{so}CustomData/{ar}KeyValueOfstringanyType/{ar}Value"
    for k, v in zip(root.findall(keys.format(**nspace)), root.findall(values.format(**nspace))):
        customDict[k.text] = v.text

    if 'AppliedDefocus' in customDict:
        acqDict['AppliedDefocus'] = float(customDict['AppliedDefocus']) * math.pow(10, 6)
    if 'Dose' in customDict:
        acqDict['Dose'] = float(customDict['Dose']) * math.pow(10, -20)
    if 'PhasePlateUsed' in customDict:
        acqDict['PhasePlateUsed'] = customDict['PhasePlateUsed']

        if customDict['PhasePlateUsed'] == 'true':
            acqDict['PhasePlateNumber'] = customDict['PhasePlateApertureName'].split(" ")[-1]
            acqDict['PhasePlatePosition'] = customDict['PhasePlatePosition']

    if 'DoseOnCamera' in customDict:
        acqDict['DoseOnCamera'] = customDict['DoseOnCamera']

    if DEBUG:
        for k, v in sorted(acqDict.items()):
            print("%s = %s" % (k, v))

    # convert all to str
    for key in acqDict:
        acqDict[key] = str(acqDict[key])

    return acqDict
